//==============================================================================
//		EAS_User.c		Program file for the EAS upgrade
//						(c) Copyright Nigel Hewitt BSc. 2006-2010
//==============================================================================

#include "C:\\Nigel\\EAS\\CPU\\EAS_Auto.h"

#define usbChar	USBAddTx		// make the built in routine conform

#define VERSION		"1.15"

//#define N_DEBUG					// put nasty things into the code

// WARNING: Despite the C compiler being case sensitive the underlying
// assembler doesn't seem to be. Hence 'RPM' and 'rpm' end up the same and when
// one is a function and the other the returned result things get confused.
//
// Also: there seems to be some confusion about intermediate results if we are
// changing type (unsigned to signed, or int to float).
// Hence I have broken a lot of maths down into stages with:
// a=b; a-=c; rather than a=b-c; and it seems far more consistant so far.
//
//==============================================================================
// USB Boot loader
// ===============
// Load the Microchip "boot loader.exe" (from their site)
// Then use the JDM jig to program the chip with
// C:\MCHPFSUSB\fw\Boot\_output\MCHPUSB.hex
// Hold down SW2 (PORTB bit 4) and reboot (SW1)
// Plug in the USB cable so windows sees it for the first time
// The driver inf file is in C:\MCHPFSUSB\Pc\MCHPUSB Driver\Release
// Run the Bootloader program C:\MCHPFSUSB\Pc\Pdfsusb\PDFSUSB.exe
// Change the blank combo box to the only selection, load the hex file
// and program. It takes about 10 seconds and when the buttons light up
// again it's done.
//
// Then you need to patch it to take out the LED drivers on PORTD bits 3 and 4
// as that means I have to unplug it every time I want to do a program upload
// but that's too big a story to go here. The fix is to replace MCHPUSB.HEX
// with boot\nig-boot.hex which is patched.
//
//==============================================================================
// Vista:
// The boot loader:
// Install the usual MCHPFSUSB folder of stuff.
// Bring up the '4550 CPU in bootload mode so the USB finds it.
// Tell it to find the driver in MCHPFSUSB\Pc and it installs.
// Make a desktop shortcut for pdfsusb.exe and set it to "XP SP1 compatability"
// mode. Then use the Control Panel/Device Manager/Other Devices/PIC 18F4550
// stuff and select the Power Management tab. Turn off the option to turn the
// device off. Then the bootloader works.
//
// Serial port emulation:
//
//==============================================================================
// Things to do to the standard library:
// in Libs\USB\CDC\C\cdc.c at line 45 add
//		#ifndef SetLineUsed				// NVH
//		#define SetLineUsed 0			// NVH
//		#endif								// NVH
//		#ifndef _SetControlUsed			// NVH
//		#define _SetControlUsed 0		// NVH
//		#endif								// NVH
// this stops a couple of warnings
//
// in Libs\USB\CDC\C\usbdrv.c at line 255
// comment out
//    if(bIDLEIF && bIDLEIE)    USBSuspend();
// this stops Microchip from putting the CPU to sleep once the USB is unplugged
//
// then in Program Files\FED\PIXIE\APPWIZ\APPWIZ18\WIZUSB.c
// at about line 19 add
//	 unsigned char nigMode=0;
// at about line 74 add nigMode to:
//   while (!nigMode && ((usb_device_state<CONFIGURED_STATE) || (bSUSPND==1)));
// this stops FED looping the CPU on us once the USB is unplugged
//
//
// No, I haven't come up with a way to prevent the occasional mad list of
// Warnings yet but thankfully it doesn't happen every time.
//
//==============================================================================
// I need some documentation for the design I put into the
// Forest Electronic Developement PIC C Compiler 
// I have ver 12.02 WIZ-C 'Full' and 'Professional' version
//
// Device 18F4550 Clock 20MHz Use Interupt=yes
// 
// USB
//		Use boot loader = yes
//		Flash USB LEDs = yes
//		Define Standard Config = yes
//		Vendor ID = 6017
//		Product ID = 2570
//		Occurrences USBByteRx = rxUSB
//		Connect pins USBLED1 = 40, USBLED2 = 39
//		You don't see pins 23 and 24 but they are connected
//		('Standard Config' is important as it sets the osc to 20MHz turns
//		 the PLL on and generally makes things run to full speed (48MHz)
//		 with the USB clocks generated correctly.)
// Capture Module 1 (adds timers 1 and 3)
//		Capture Falling Edge
//		Occurences Capt1  = fmph
//		Connect pins CCP1 = 17 = ROAD_SPEED
// Capture Module 2
//		Capture Falling Edge
//		Occurences Capt2  = frpm
//		Connect pins CCP1 = 16 = ENGINE_SPEED
// USART
//		Bit rate = 9600
//		Occurrences RxByte = rxSerial
//		Connect pins Tx = 25, Rx = 26
// Data EEPROM
//		Occurrences EEFin = eeDone
// A/D converter 18x
//		AFR=2, AFL=3 ARR=4, ARL=7
//		Enable = yes
//		Initial channel 0
//		Source for A/D OscDiv32
//		Number of A/D channels 5
//		Read as 10 bit lo-hi = yes
//		Acquisition Time TAd4
//		Source for +ve Vref_Vdd
//		Source for -ve Vref_Vss
//		Occurrences ADDone = adDone
// I was a little concerned about the way it allocates a pin to A 'OR' D in the
// set up specification but, as it does warn you to set all pins in the TRIS
// register to inputs 'or it is the digital output voltage that is measured',
// I suspect that outputs LED1 and LED2 (AN3 and 4) will work consistantly.
// Outputs
//		LED1		 = 5  initial = 0
//		LED2		 = 6  initial = 0
//		SRF		 = 19 initial = 0
//		SRL		 = 20 initial = 0
//		SRR		 = 21 initial = 0
//		SLR		 = 22 initial = 0
//		SEXH		 = 27 initial = 0
//		SINP		 = 28 initial = 0
//		RHSW_DATA = 33 initial = 1
//		RHSW_CLK	 = 34 initial = 1
//		COMPPWR	 = 38 initial = 0
// Simple Switch (adds timer 0)
// 		all: delay before repeat = 500
//			 delay between repeats	 = 250
//			 debounce					 = 60
//			 Enable repeat				 = no
//			 Use pull-down				 = no
//		UP			 = 8		SBPress0 = up   
//		DOWN		 = 9		SBPress1 = down
//		INHIBIT	 = 10		SBPress2 = hold
//		DOORS		 = 15		SBPress3 = doors
//		OVERHEAT	 = 29
//		PRESSURE	 = 30
//		BRAKES	 = 35
//		HANDBRAKE = 36
//		SENSE		 = 37   SBPress8 = Sense
// Timers 0, 1 and 3 are automaticaly added
// Adjust Timer 1 to
//		Enable Timer 1					yes
//		Use internal clock source	yes
//		Synchronise Timer 1 input	yes
//		Enable Timer 1 oscillator	no
//		Timer 1 prescaler				8
//		Timer 1 read is 16 bit		no
// Timer 3 is not used and is only inserted because one of its bits
// sets the capture ports to have their timing source set to T1_SrcBoth.
// I use it to generate the 1mS heart beat clock.
// NB: I am not interupting so the service routine just runs at the next
//	   opertunity but it seems not to jitter significantly.
// Timer T3
//		Enable Timer 3				yes
//		Use external clock		no
//		Do not synchronise...	no
//		Timer 3 prescaler			8
//		Source for Capture/Compare	T1_SrcBoth
//		Enable 16 bit r/w			no
//
//		occurance T3_of	= mSec
//
//==============================================================================
// Time to name and describe the inputs/outputs
// Inputs
// Name			CPU	Plug	Desc
// UP				8	32		0 = UP button pressed
// DOWN			9	33		0 = DOWN button pressed
// INHIBIT		10	15		0 = Inhibit switch is pressed
// DOORS			15	34		1 = doors closed
// OVERHEAT		29	16		0 = ok
// PRESSURE		30	13		1 = up to pressure
// BRAKES		35	31		1 = footbrake pressed
// HANDBRAKE	36	14		1 = Handbrake released AND in gear
// SENSE			37	n/a	0 = SW2 pressed
//
// Outputs
// SRF			19	27		1 = power to front right solonoid
// SLF			20	10		1 = power to front left solonoid
// SRR			21	28		1 = power to rear right solonoid
// SLR			22	11		1 = power to rear left solonoid
// SEXH			27	9		1 = power to outlet valve
// SINP			28	26		1 = power to inlet valve
// COMP			38	8		1 = power to compressor relay
//
// Odds
// ES			16	12		engine speed pulse train (spark rate so 4 per rev)
// RS			17	30		road speed pulse train....
// RHCLK		34	25		ride height switch clock
// RHDATA	33	7		ride height switch data
// TX			25	35		serial transmit to testbook
// RX			26	17		serial receive to tesbook
//
//------------------------------------------------------------------------------
// shorthands
//------------------------------------------------------------------------------

union WD { unsigned int w; unsigned char b[2]; };	// b[0]=lsbyte

#define SIZEOF(A)	(sizeof(A)/sizeof(A[0]))	// size of array in elements
//#define OFFSET(S,I) ((int)&(((S*)0)->I))		// offset of item I in struct S
#define abs(A)	((A)>=0?(A):-(A))				// absolute value

void compressorTick();			// Protect forward references
void eeWrite();
void usbText(char *p);

//------------------------------------------------------------------------------
// global variables
//------------------------------------------------------------------------------

// The variables are in a named structure or otherwize they are not allocated
// in the order specified and a simple block copy up the USB for diagnostic
// and monitoring purposes doesn't work. (also some adjustment to the packing
// is required but see the VC++ code for that)

#define short		// this allows me to cut and stick the union into VC++
					// and the bits all line up
struct {
	
unsigned char RAM_START;			// 'S'

// debug hold via SW2
unsigned char bHold;

// Time values
unsigned short int mSecs;			// 0-999
unsigned short int secs;			// 0 to 18 hours

// A2D values for suspension
// Readings are taken every tenth of a second
// ports are ( 0,1,2,4 )=( RF,LF,RR,LR )
unsigned char adN;					// current A2D channel index 0-3
unsigned char adIndex;				// index to averages (points to most recent)
union WD	  adRaw[4];					// current 'raw' values 0-1023 = 0-5V
#define AD_MAX	900					// out of range tests (about 930 when n/c)
#define AD_MIN	300
unsigned short int adFinal[4];	// processed height in mm
// Then we need an averaged version for when we are on the move
// I choose to do readings every tenth of a second and keep 2 seconds worth
#define NSLOTS	20						// 10 per second for 2 seconds
unsigned short int hLag[NSLOTS][4];
// These are then processed and normalised to give the final version
unsigned short int height[4];		// translated values in mm
#define hRF	x.height[0]				// and a few shorthands
#define hLF x.height[1]
#define hRR x.height[2]
#define hLR x.height[3]
unsigned char nLag;					// slot pointer

// Pulse values
unsigned char rpmTicks;				// RPM service counter
unsigned short int prevRPM;		// previous counter
unsigned short int rawRPM;			// RPM counts per pulse
unsigned short int RPM;				// processed RPM
unsigned char mphTicks;				// MPH counts per second counter
unsigned short int rawMPH;			// MPH counts per second
unsigned short int MPH;				// processed MPH

// Compressor operation
unsigned char compON;				// we are compressing (0=no, 1=yes, 2=cooling)
unsigned short int compTimer;		// pump too long = leak
#define COMP_LIMIT	600			// ten minutes at 10 counts per second
#define COMP_RESET	300			// give it five minutes to cool down

// Command system
unsigned char usbCmd[19];			// command line
unsigned char usbNcmd;				// number of bytes in the command line

// Ride height switch system
// These are the 'official' values.
// If xBUGS or xINHIBIT are set I may well do my own thing
unsigned char rhLights;				// 1=HIGH 2=NORM 4=MWAY 8=ACCESS
unsigned char rhFlash;				// as above but flashing
unsigned char rhAlt;					// half second flag to do the flash
unsigned char rhWork;				// working values

// 'high speed' logic
unsigned char bHighSpeed;			// logic flag
unsigned char nHighSpeed;			// seconds counter

// Mode selection and transition
// We have officially three selectable modes, I add two more
enum UMODE { M_BUMPSTOPS, M_ACCESS, M_NORMAL, M_HIGH, M_VERYHIGH };
unsigned char uMode;					// =M_NORMAL desired mode
// However this translates into a more complex mode when you add the rest
enum SMODE { S_BUMPSTOPS, S_ACCESS, S_LOW, S_NORMAL, S_HIGH,
			 S_VERYHIGH, S_UNKNOWN };
unsigned char sMode;					// =S_UNKNOWN  current mode
unsigned char tMode;					// =S_NORMAL   target mode
unsigned char tHeight;				// target height
// and we need a set of proposed heights for each mode
unsigned char mHeight[6];
//	0,			// Bumpstops
//	10,		// Access				-2.5"  (-63mm)
//	48,		// Low Profile			-1.0"  (-25mm)
//	73,		// Standard Profile	 0"	 ( 0mm)
//	112,		// High Profile		+1.5"  (+38mm)
//	130		// Extended Profile	+2.25" (+57mm)
//	150		// This is where we cap values

enum { M_HOLD,														// no planned move
	   M_URF, M_ULF, M_URR, M_ULR, M_UF, M_UR,			// up moves
	   M_DRF, M_DLF, M_DRR, M_DLR, M_DF, M_DR, M_D,		// down moves
	   M_SETTLE=0x80 };											// apres move dwell
unsigned char mState;				// what we are doing
unsigned char tState;				// state working timer (tenths of a second)
unsigned char step;					// step change so beware of dazzle
// To access the extra modes we need some button hold counters...
unsigned char upHold;
unsigned char downHold;

// eeWrite system
unsigned short int eeAddr;			// counter for bytes written

// The bug system
// Any bit set here is a catastrophic bug and we have to stop running things.
// However notice that in my world switching off and on again gives things a
// second chance

unsigned short int bugs;				// 16 bits
// these are the bug bits
#define bugOVERHEAT		0				// overheat flag set
#define bugOVERRUN		1				// compressor ran for too long
#define bugSTRUTHI		2				// over volts for 4 struts
#define bugSTRUTLO		6				// under volts for 4 struts
#define bugTOOHIGH		10				// one strut too high (debug)
#define bugTOOLOW			11				// one strut too low  (debug)
#define bugFATAL			0  			//0xFFFF	// for now all bugs are fatal


// Constants block we read from the eerom
unsigned char eeRom_errors[16];			// error flags
unsigned short int eeRom_offsets[4];	// bump stop height offsets
#define EECOUNT	(sizeof(x.eeRom_errors)+sizeof(x.eeRom_offsets))
#define ERROR_COUNT sizeof(x.eeRom_errors)

// For debugging I need some reported slots so I don't have to recompile
// the windows program in the car.
unsigned short int vDebug1;
unsigned short int vDebug2;
unsigned short int vDebug3;
unsigned short int vDebug4;
unsigned short int nDump;				// memory dump counter

// this is used to mask the output ports
unsigned char _ports[5];				// ={ 0,3,0,0,0 };
unsigned char _inputs[5];				// ports as read

unsigned char RAM_END;					// 'F'
}x;

// Definitions for output bits as port,bit
#define _PORTA	0							// my version of the ports
#define _PORTB	1
#define _PORTC	2
#define _PORTD	3
#define _PORTE	4

#define xLED1		_PORTA,3
#define xLED2		_PORTA,4
#define xSRF		_PORTD,0
#define xSLF		_PORTD,1
#define xSRR		_PORTD,2
#define xSLR		_PORTD,3
#define xSEXH		_PORTD,4
#define xSINP		_PORTD,5
#define xRHDATA	_PORTB,0
#define xRHCLK		_PORTB,1
#define xCOMP		_PORTB,5

// The inputs are a bit funny. Simply using them as logicals, as in the code
// examples may work in as if(X) but not in X && Y constructions.
// These are my translations
#define xUP			(UP==0)						// buttons pressed
#define xDOWN		(DOWN==0)
#define xINHIBIT	(INHIBIT==0)
#define xDOORS		(DOORS==0)					// doors open
#define xOVERHEAT	(OVERHEAT!=0)
#define xPRESSURE	(PRESSURE!=0)				// up to required pressure
#define xBRAKES	(BRAKES!=0)					// brakes applied
#define xROLL		(HANDBRAKE!=0)				// handbrake released and in gear
#define xSW2		(SENSE==0)
#define xBUGS		((x.bugs&bugFATAL)!=0)	// a dummy but working the same way
#define xENGINE	(x.RPM>200)					// tick-over is about 500+

//------------------------------------------------------------------------------
// Writing ports
//------------------------------------------------------------------------------
// There must be something very stupid I'm doing but my port bits won't stay
// written. I set a bit and the next time something writes to that PORT it gets
// cleared. I thought it was just PORTA where I am messing with the A2D
// converter but it happens elsewhere.
// This may be overkill but it is a fix

// An array representing ports A to E
// all initialised to zero except B0 and B1
//unsigned char x._ports[5]={ 0,3,0,0,0 };

// The ports are defined as PORTA  0x0F80 and sequentially

void out(unsigned char pt, unsigned char bt, unsigned char val)
{
	if(val!=0)	x._ports[pt] |= 1<<bt;
	else		   x._ports[pt] &= ~(1<<bt);
	*(&PORTA+pt) = x._ports[pt];
}
void flip(unsigned char pt, unsigned char bt)
{
	x._ports[pt] ^= 1<<bt;
	*(&PORTA+pt) = x._ports[pt];
}
unsigned char get(unsigned char pt, unsigned char bt)
{
	return (x._ports[pt]>>bt) & 1;
}
//------------------------------------------------------------------------------
// Setting a bug - not all bugs need be fatal, see bugFATAL
//------------------------------------------------------------------------------

void bug(unsigned char b)
{
	if(b>16) return;					// safety trap
	x.bugs |= 1<<b;
	x.eeRom_errors[b] = 1;
	// since these are one byte items we can write them directly
	WriteEEData(b, x.eeRom_errors[b]);
}
//------------------------------------------------------------------------------
// Initialisation
//------------------------------------------------------------------------------
// All the output pins (other than Ride Height Switch bits) are set to 0 as an
// initial value so it should start soft.

// This is needed to work the fix in the USB code to prevent the loopup when
// the USB cable is unplugged. See notes in first section
extern unsigned char nigMode;

void UserInitialise()
{
	nigMode = 1;

	// clear the memory block
	unsigned char *p=&x.RAM_START;
	unsigned char *q=&x.RAM_END;
	while(p<=q) *p++=0;
		
	// do the presets...
	x.RAM_START		= 'S';
	x.uMode			= M_NORMAL;
	x.sMode			= S_UNKNOWN;	// current mode
	x.tMode			= S_NORMAL;		// target mode
	x.mHeight[0]	= 0;				// Bumpstops
	x.mHeight[1]	= 10;				// Access				-2.5"  (-63mm)
	x.mHeight[2]	= 48;				// Low Profile			-1.0"  (-25mm)
	x.mHeight[3]	= 73;				// Standard Profile	 0"	   ( 0mm)
	x.mHeight[4]	= 112;			// High Profile		+1.5"  (+38mm)
	x.mHeight[5]	= 130;			// Extended Profile	+2.25" (+57mm)
	x.mState			= M_HOLD;		// what we are doing
	x._ports[1]		= 3;
	x.RAM_END		= 'F';

	// Read the eerom data
	p = (unsigned char*)&x.eeRom_errors[0];
	for(unsigned char i=0; i<EECOUNT; i++)
		p[i] = ReadEEData(i);

	// if the eeRom is not set
//	if(eeRom.offsets[0]==0 || eeRom.offsets[0]=0xffff){
		// or 806, 790, 769 and 792 on 2006-11-26
		// or 802, 808, 776 and 793 on  first test
		// or 782, 775, 775 and 772 on 2008-04-06
		x.eeRom_offsets[0] = 782;	// results from first test
		x.eeRom_offsets[1] = 775;
		x.eeRom_offsets[2] = 775;
		x.eeRom_offsets[3] = 772;
//		eeWrite();						// start a write cycle to put them in
//	}
	
	// Dummy up some heights until the lag settles
	x.height[0] = x.height[1] = x.height[2] = x.height[3]
		 = x.tHeight = x.mHeight[S_NORMAL];		 
}
//------------------------------------------------------------------------------
// Interupt main vector
//------------------------------------------------------------------------------

void UserInterrupt()
{
 #asmline goto UserIntReturn	; PIC Assembler - go back to interrupt routine
}
void die()
{
 #asmline goto RealStart
}
//------------------------------------------------------------------------------
// A2D converter has finished a conversion
//------------------------------------------------------------------------------
// Datasheet page 261
// I told the A2D setup it had 5 channels, we actually use 0,1,2 and 4
//
// We need to calculate the data acquisition time but it is going to be less
// than their example of 2.5uS. (page 266)
// We need to set the A2D clock to 0.7 to 25uS (page 400) with advice to keep
// it short. Clock=20MHz=Tosc=50nS so divide by 32 gives 1.6uS.
// Then make the acquisition time 4Tads 6.4uS which should be pleanty.
// Those numbers will also survive me jacking the clock to 48MHz later if
// required. (done at 1.05)
// Thse selections are made on the 'Application Designer' page and I have
// edited them into the notes at the begining of this file.
// This means we take 6.4uS to settle then 19.2uS to convert so all 4 results
// will be done in 103uS. To prevent us spending all our time servicing the
// A2D interupts for values I can't use I shall poll them by starting a new
// conversion every quarter of a second.
//
// The sensors are supplied with 6.8V but due to the amplifiers we measure
//		0 counts = 5.2V and 1023 counts at about 0V
//
// We are expecting values in the region of 800 (low) to 400 (high)
// representing about 4 inches of travel. I am going to choose to work in
// millimeters giving 4 inches as about 100 with zero being the bump stops
// We will provide a calibration system where you tell it to calibrate and it
// drops itself, measures the settings and records them.
//
// NB: the sensors are ordered RF,LF,RR,LR
// Measuring from the top of the hub boss to the highest 'bottom' of wheel arch
// gives me 36, 35, 37 and 35cms so I take those as my 'zero' points.
// the raw height readbacks at this point were: 326H 316H 301H 318H so
// 806, 790, 769 and 792. A further set of readings were
// 2CAH 259H 29EH 22DH => 714, 601, 670, 557 for 38 43 42 and 46cms
// 92counts=2cms, 189=8, 99=5 and 235=11 so 23.65 counts per cm
// 2.365 counts/mm so *74/175 to get mm
// I have target heights up to 130 so I will treat >150 (5.9inches) as an error


void adDone()   
{
	static unsigned char r[4]={ 0,1,2,4 };	// lookup to manage the ports
	x.adRaw[x.adN].b[0] = ADRESL;			// unpack the result
	x.adRaw[x.adN].b[1] = ADRESH&3;			// 10 bits (0-1023)

	// first trap the errors before we do something very silly
	if(x.adRaw[x.adN].w>AD_MAX && x.secs>6)
		bug(bugSTRUTHI+x.adN);
	else if(x.adRaw[x.adN].w<AD_MIN && x.secs>6)
		bug(bugSTRUTLO+x.adN);
	// continue processing so we can debug via the USB
	
	// the adRaw[].w values are the 'raw' height values output
	// adFinal[] are the 'processed' values
	
	// the offset is about 850 and represents fully down
	// 400 represents fully up
	// apply the offsets
	
	if(x.adRaw[x.adN].w>x.eeRom_offsets[x.adN]){  // ie beyond the 'zero' point
		x.adFinal[x.adN] = 0;
		if(x.secs>6)
			bug(bugTOOLOW);		// testing getout clause
	}
	else{
		// I'm having trouble with the convert.
		// I suspect compiler problems so the maths is broken down into steps.
		// It works a lot better this way so I assume there is a problem
		// deciding what type to put the results in
		int t1 = x.eeRom_offsets[x.adN];
		t1 = t1 - x.adRaw[x.adN].w;
		double t2 = t1;
		t2 = t2 * 0.422;
		x.adFinal[x.adN] = t2;
		if(x.adFinal[x.adN]>150){
			if(x.secs>6)
				bug(bugTOOHIGH);		// testing getout clause
			x.adFinal[x.adN] = 150;
		}
	}
	// cycle to the next strut
	if(++x.adN<4){						// next port
		ADCON0 = (r[x.adN]<<2) | 3;		// trigger the next read
		return;
	}
	// we get here when we finish processing the fourth strut
	x.adN = 0;							// reset for next time

	// We have now generated the four basic height values and we need to
	// provide a lagged system to even out the bumps when we are moving
	if(++x.nLag>=NSLOTS) x.nLag=0;
	x.hLag[x.nLag][0] = x.adFinal[0];
	x.hLag[x.nLag][1] = x.adFinal[1];
	x.hLag[x.nLag][2] = x.adFinal[2];
	x.hLag[x.nLag][3] = x.adFinal[3];
}
// Calibration is the simple task of dropping to the bump stops and copying the
// current adRaw[] values to the offsets and saving them to the eRom
// My first measured offsets were 802, 808, 776 and 793

void calibrate()
{
	x.eeRom_offsets[0] = x.adRaw[0].w;
	x.eeRom_offsets[1] = x.adRaw[1].w;
	x.eeRom_offsets[2] = x.adRaw[2].w;
	x.eeRom_offsets[3] = x.adRaw[3].w;
	eeWrite();
}
// The lagged height values have to be processed into the final height[] array
// the array was initialised to the 'normal' heights.
// I break this out from the A2D stuff as you only need to do the averaging
// maths when you actually need the values.

void normalise()
{
	if(x.secs>3){							// let things settle on switch on
		if(xROLL && xENGINE){				// if moving use averaging
			for(int i=0; i<4; i++){
				int k=0;
				// since max height is 150 we can sum in an integer
				// until we want NSLOTS>218
				for(int j=0; j<NSLOTS; j++)
					k += x.hLag[j][i];		// sum and...
				x.height[i] = k/NSLOTS;		// ...average
			}					
		}
		else{
			for(int i=0; i<4; i++)
				x.height[i] = x.adFinal[i];
		}
		// From now on we can refer to the hRF,hLR etc. macros...
		// Now in a world of perfect sensors and billiard table flat roads
		// we would be finished but we need height numbers that work.
		// The rule for a rigid rectangular chassis is that the two diagonals
		// must coincide so hRF+hLR = hLF+hRR
//		int err = ((int)(hRF+hLR) - (int)(hLF+hRR))/4;
//		if(err>hRF)  err=hRF;	// limit so we don't push anybody negative
//		if(err>hLR)  err=hLR;
//		if(-err>hLF) err=-hLF;
//		if(-err>hRR) err=-hRR;
//		hRF -= err;
//		hLR -= err;
//		hLF += err;
//		hRR += err;
		// Of course if one sensor was OOR we could infer its reading rather
		// than correct out the error but I don't think I want to go down that
		// path just yet.
	}
}
//------------------------------------------------------------------------------
// Data received on the USB port
//------------------------------------------------------------------------------
// Datasheet page 165
// NB: the software USB buffer is 64 bytes
// I am using the Combro's Vendor=6017 Product=2570 so it signs on as a
// ballistic chronoscope.

void rxUSB()   
{
	unsigned int n = USBGetRx();		// read the char
	x.usbCmd[x.usbNcmd] = n;			// always buffer it
	x.usbCmd[x.usbNcmd+1] = 0;			// ensure there is a null
	if(x.usbNcmd<sizeof(x.usbCmd)-2)	// but only increment if there is room
		x.usbNcmd++;
//	USBAddTx(n);						// echo
}
//------------------------------------------------------------------------------
// eerom write has finished. (this takes too long to loop waiting for it)
//------------------------------------------------------------------------------

// unsigned char ReadEEData(unsigned char Address);
// void WriteEEData(unsigned char Address, unsigned char Data);
// the write can take 5 to 10mS so we get called back on finish
// Don't call this if eeAddr!=0

// this is our call to start the service
void eeWrite()
{
	x.eeAddr = 0;
	WriteEEData(x.eeAddr, ((char*)x.eeRom)[x.eeAddr]);
	x.eeAddr++;
}
// end of write cycle
void eeDone()   
{
	if(x.eeAddr && x.eeAddr<EECOUNT){
		WriteEEData(x.eeAddr, ((char*)x.eeRom)[x.eeAddr]);
		x.eeAddr++;
	}
	x.eeAddr = 0;		// clear and ready for next time
}
//------------------------------------------------------------------------------
// Value ready at the CCP1 port representing MPH
//------------------------------------------------------------------------------
// Datasheet page 143
// I have both CCPs set to use T1 and T1 is set to prescale by 8 on
// the master CLK by 4. So 48MHz becomes 1.5MHz on which one count represents
// 666.6nS and 0xffff counts is 43.7mSecs. This does mean we can't count times
// below 22Hz.
// Sad that.
// We only need a rough speed for our calculations so if we drop the fancy
// capture stuff and just count the ticks in a second. You will find the rest
// of the works in the mSec() function.

void fmph()   
{
	x.mphTicks++;	
}
//------------------------------------------------------------------------------
// Value ready at the CCP2 port representing RPM
//------------------------------------------------------------------------------

// The pulses seem to be at spark rate so
// 3K rpm would 50rps so 4 pulses per rev (spark rate) would be 200pps
// 200pps is 5mSecs per period so on the 1.5MHz clock 7500 counts.
// hence rpm = 22500000/N is the formula.
// Tested with:
// idle=31140(722) 1k=21580(1042) 2K=11148(2018) 3K=7600(2960) 4K=5600(4017)
//
// Since we don't get called when the engine if off we clear a watchdog
// that is monitored by the mSec tick. At idle we should be ticking at 25mS
// intervals so if the counter goes above 50mS then RPM is zeroed.
// As the first count after that will be trash if RPM==0 we set it to 1 and
// go again.

void frpm()   
{
	union WD v;
	v.b[1] = CCPR2H;			// get the value
	v.b[0] = CCPR2L;
	
	x.rpmTicks = 0;					// reset the watchdog
	x.rawRPM   = v.w-x.prevRPM;	// calculate the delay (unsigned maths wraps)
	x.prevRPM  = v.w;

	if(x.RPM==0) x.RPM = 1;
	else{
		long d=22500000L;			// it doesn't do literals very well
		d = d/x.rawRPM;			// it certainly doesn't do long divides either
		x.RPM = d;					// but this works
	}
}
//------------------------------------------------------------------------------
// Data received on the serial port
//------------------------------------------------------------------------------
// For now we just echo

void rxSerial()   
{
	if(PIR1&(1<<TXIF))	// if txNotBuzy
		TXREG=RCREG;	// copy rxIn to txOut
}
//------------------------------------------------------------------------------
// UP, DOWN and INHIBIT switches (debounced and not repeating)
//------------------------------------------------------------------------------
// This routine can't select the extreme modes so there are bits in mSec() and
// also tilt() to sense the press-and-hold operations

void up()
{
	// if not INHIBIT and settled change the user mode
	if(!xINHIBIT && !xBUGS && !x.bHold){
		if(x.uMode==M_BUMPSTOPS)	x.uMode = M_ACCESS;
		else if(x.uMode==M_ACCESS)	x.uMode = M_NORMAL;
		else if(x.uMode==M_NORMAL)	x.uMode = M_HIGH;
		
		if(x.sMode!=S_UNKNOWN) x.step=true;
	}
}
void down()
{
	if(!xINHIBIT && !xBUGS && !x.bHold){
		if(x.uMode==M_VERYHIGH)		x.uMode = M_HIGH;
		else if(x.uMode==M_HIGH)	x.uMode = M_NORMAL;
		else if(x.uMode==M_NORMAL)	x.uMode = M_ACCESS;
		
		if(x.sMode!=S_UNKNOWN) x.step=true;
	}
}
void hold()
{
	// once xINHIBIT sets we won't write to these so the debugger can play
	out(xSRF, 0);
	out(xSLF, 0);
	out(xSRR, 0);
	out(xSLR, 0);
	out(xSEXH, 0);
	out(xSINP, 0);

	// hold UP and press INHIBIT to clear current bugs
	if(xUP){
		x.bugs = 0;
	}
	// hold DOWN and press INHIBIT to calibrate
	else if(xDOWN && x.sMode==S_BUMPSTOPS && x.tMode==S_BUMPSTOPS)
		calibrate();
}
void doors()	// this routine is also used to shut everything off
{
	out(xSRF, 0);
	out(xSLF, 0);
	out(xSRR, 0);
	out(xSLR, 0);
	out(xSEXH, 0);
	out(xSINP, 0);
}
void Sense()	// SW2 pressed
{
	x.bHold = !x.bHold;
	usbText(x.bHold ? "HOLD\r\n" : "RUN\r\n");
}
//------------------------------------------------------------------------------
// Ride height switch lights system
//------------------------------------------------------------------------------
// called every mSec

void rhSwitch()
{
	// Now run the Ride Height Switch
	// See the picture C:\Nigel\EAS\rhsw3.jpg
	// The lines are inverted coming out from the CPU but the switch uses high
	// for on and a rising edge trigger.
	// The output frame has two slightly different forms and cycles every 100mS
	// so I will treat it as a 200mS total cycle.
	// the clock goes 'low' for 2mS and is 'high' for 8mS four times inverted
	
	// Try this from the rangerover.net forums:
	// To clear the "EAS Fault, slow 35mph max" warning from the dash, put two
	// jumpers on the connector to the ECU. one from pin 7 to pin 18 (ground )
	// and the other from pin 25 to pin 1 (+12v). When you startup you will get
	// a soft beep and short "EAS manual" message, then normal message center
	// operation will resume.
	// Well Pin 7 is xRDDATA and pin 25 is xRHCLK and their output circuits
	// perform an invert so 1 sets GND and 0 sets 12V

	if(x.bHold){							// if we are locked in USB mode...
		out(xRHDATA, 1);
		out(xRHCLK, 0);
		return;
	}
	
	x.rhAlt = x.mSecs>=500;					// generate the flash

	switch(x.mSecs%200){
	case 0:
	case 100:
		x.rhWork = x.rhLights;				// 1=HIGH 2=NORM 4=MWAY 8=ACCESS
		if(x.rhAlt) x.rhWork ^= x.rhFlash;	// xor so even an 'on' light flashs
		out(xRHCLK, 1);						// clock 'low'
		out(xRHDATA, (x.rhWork & 8)==0);	// data for bottom light inverted
		break;
	case 2:   case 12:  case 22:  case 32:
	case 102: case 112: case 122: case 132:
		out(xRHCLK, 0);						// clock to high
		break;
	case 4: case 14: case 24:
	case 104: case 114: case 124: // case 134:
		out(xRHDATA, 0);					// return data to high
		break;
	case 34: case 134:
		out(xRHDATA, 1);					// alternating data low for the gap
		break;
	case 10:
	case 110:
		out(xRHCLK, 1);						// clock 'low'
		out(xRHDATA, (x.rhWork & 4)==0);	// data for next light inverted
		break;
	case 20:
	case 120:
		out(xRHCLK, 1);						// clock 'low'
		out(xRHDATA, (x.rhWork & 2)==0);	// data for normal light inverted
		break;
	case 30:
	case 130:
		out(xRHCLK, 1);						// clock 'low'
		out(xRHDATA, (x.rhWork & 1)==0);	// data for top light inverted
		break;
	}
}
//------------------------------------------------------------------------------
// Compressor management
//------------------------------------------------------------------------------

// Considerations:
// We pump too long and the pressure switch doesn't set - we have a leak.
// We get asked to pump but we're not using gas - we have a smaller leak.
// We try to lift a leg and go OOG but no lift - we have a big leak

// I started with a complicated vision of the compressor logic but in fact it
// all comes down to one simple logic equation

void compressorTick()
{
	// Run even with inhibit pressed so we have gas to play with
	if(x.bHold) return;
	// compON is 0 for off, 1 for on and 2 for cooling down
	// if it is 2 we mustn't touch it until the counter clears
	if(x.compON!=2)
		x.compON = !xBUGS					// no show stoppers
			  && xENGINE				// engine running
			  && !get(xSEXH)			// not dumping_gas
			  && !xPRESSURE			// below_pressure
			  && !xOVERHEAT;			// not overheat
	//		  && x.compTimer<COMP_LIMIT;

	// in this initial version be pretty brutal about problems
	if(xOVERHEAT) bug(bugOVERHEAT);
//	if(x.compTimer>=COMP_LIMIT) bug(bugOVERRUN);
	
	out(xCOMP, x.compON==1);
}
//------------------------------------------------------------------------------
// Self leveling algorythm
//------------------------------------------------------------------------------
// This gets called 10 times a second

// Start with a couple of workers...
// use RF=1<<0, LF=1<<1, RR=1<<2, LR=1<<3

void pump(unsigned char n)
{
	out(xSRF,  n&1);
	out(xSLF,  n&2);
	out(xSRR,  n&4);
	out(xSLR,  n&8);
	out(xSEXH, 0);
	out(xSINP, n&15);
}
void dump(unsigned char n)
{
	out(xSRF,  n&1);
	out(xSLF,  n&2);
	out(xSRR,  n&4);
	out(xSLR,  n&8);
	out(xSEXH, n&15);
	out(xSINP, 0);
}
void tilt()
{
	// Let's start with the 'short way' with INHIBIT switches
	// both INHIBIT and DOORS shut everything off when they are actioned
	// so if I do no processing I can use the manual USB controls
	if(xINHIBIT || xDOORS || x.bHold){
		x.mState = M_HOLD;
		x.tState = 0;
		return;				// this saves me testing for INHIBIT/etc. elsewhere
	}
	if(xBUGS){				// keep this seperate for now
		doors();			// shut everything off
		x.mState = M_HOLD;
		x.tState = 0;
		return;
	}
	// Convert the requested mode plus other things into a target mode
	// uMode (user mode) is:
	// enum UMODE { M_BUMPSTOPS, M_ACCESS, M_NORMAL, M_HIGH, M_VERYHIGH };
	// sMode and tMode (current and target modes) are:
	// enum SMODE { S_BUMPSTOPS, S_ACCESS, S_LOW, S_NORMAL, S_HIGH,
	//				S_VERYHIGH, S_UNKNOWN };

	switch(x.uMode){		// requested mode from rhswitch (M_ values)
	case M_BUMPSTOPS:
 bs:	if(x.MPH<5)				x.tMode = S_BUMPSTOPS;
		else if(x.MPH<20)		x.tMode = S_ACCESS;
		else					x.tMode = S_LOW;
		break;
	case M_ACCESS:
		if(x.MPH>20)			x.tMode = S_LOW;
		else if(x.downHold>50){	x.uMode = M_BUMPSTOPS; goto bs; }
		else					x.tMode = S_ACCESS; 
		break;
	case M_NORMAL:
	default:
		if(x.bHighSpeed)		x.tMode = S_LOW;
		else			 		x.tMode = S_NORMAL;
		break;
	case M_HIGH;
		if(x.MPH>35)			x.tMode = S_NORMAL;
		else if(x.upHold>50){	x.uMode = M_VERYHIGH; goto vh; }
//		else if(x.bStuck)		x.tMode = S_VERYHIGH;
		else					x.tMode = S_HIGH;
		break;
	case M_VERYHIGH:
 vh:	if(x.MPH<5)				x.tMode = S_VERYHIGH;
		else if(x.MPH<35)		x.tMode = S_HIGH;
		else					x.tMode = S_NORMAL;
		break;
	}
	// OK so that gives us a target mode so we now have a target height
	x.tHeight = x.mHeight[x.tMode];
	
	// Now work out the lights based on the current selected mode
	// The current state remains lit and the target flashes for normal modes
	// 1=HIGH 2=NORM 4=LOW 8=ACCESS
	x.rhLights = x.rhFlash = 0;
	switch(x.sMode){
	case S_BUMPSTOPS:	x.rhLights = 12;	break;		// bottom two lights
	case S_ACCESS:		x.rhLights = 8;		break;
	case S_LOW:			x.rhLights = 4;		break;
	case S_NORMAL:		x.rhLights = 2;		break;
	case S_HIGH:		x.rhLights = 1;		break;
	case S_VERYHIGH:	x.rhLights = 3;		break;		// top two lights
//	default:
//	case S_UNKNOWN:		x.rhLights = 0;		break;		// off if unknown
	}
	if(x.sMode!=x.tMode)					// if target mode is different...
		switch(x.tMode){
		case S_BUMPSTOPS:	x.rhFlash = 3;	break;		// confused mode 1
		case S_ACCESS:		x.rhFlash = 8;	break;
		case S_LOW:			x.rhFlash = 4;	break;
		case S_NORMAL:		x.rhFlash = 2;	break;
		case S_HIGH:		x.rhFlash = 1;	break;
		case S_VERYHIGH:	x.rhFlash = 12;	break;		// confused mode 2
		}

	// Well the height control is based on making plans, implementing them
	// and then working with what happens so it is a minimalist state machine.
	// We set a set of values on the solonoids, let it run and then move on.
	
	// If we left something running then let it run
	if(x.tState && --x.tState)		// count down in tenths of a second
		return;

	// If we get to here the previous state just timed out
	pump(0);						// beyond here we are deciding what to do
	if(x.mState&M_SETTLE)			// were we settling?
		x.mState = M_HOLD;			// yes, so now we're thinking about it
	else if(x.mState!=M_HOLD){		// were we moving but not settling?
		x.tState = xROLL ? 20 : 3;	// do some settling time
		x.mState |= M_SETTLE;		// this way means the log shows
									//		what we are settling from
		return;
	}
	// If we get here the state is M_HOLD and we are ready to recalculate

	// The big constants that make things work
	#define H_NEAR		5			// who cares about 5mm?
	#define H_RATE		3			// mm per tenth of a second
	// IMPORTANT: H_NEAR > H_RATE or we might develop some 0 time periods
	
	// also RF=0, LF=1, RR=2, LR=3 in all the arrays
	// this actually means the 0x01=left, 0x02=back
	
	// Do the easy one first...
	if(x.tMode==S_BUMPSTOPS){
		if(x.sMode!=x.tMode){
			dump(15);			// everybody down
			x.mState = M_D;
			x.tState = 100;		// give it ten seconds
			x.sMode  = x.tMode;
		}
		return;
	}
	// We now need the measured height values so...
	normalise();		// ...only do the averaging when we need it

	// Before we try anything check the engine. We get started up at odd times
	// to do a 'level' but if we try a 'top up' we will just bleed down the
	// reseviour and it will take ages to pump up again.
	bool bAllowUp = xENGINE;	// up only if the engine is running
	
	// The first question is are we doing a 'step'?
	// ie: were we OK and now we have a change to make?
	// To prevent giving the headlights to the oncoming cars the idea is that
	// the front goes down first or the rear goes up first.
	if(x.step){
		// we can't step if things are too different
		if(abs(hRF-hLF)>H_NEAR*2) x.step = false;
		if(abs(hRR-hLR)>H_NEAR*2) x.step = false;
	}
	if(x.step){
		int hFront = (hRF+hLF)/2;			// average front
		int hRear  = (hRR+hLR)/2;			// average rear
		int eFront = x.tHeight - hFront;	// positive requires an 'up'
		int eRear  = x.tHeight - hRear;
		// process nose-down and tail-up first
		if(eFront<-H_NEAR){					// if the front requires a down...
			x.tState = -eFront/H_RATE;		// tenths of a second
			x.mState = M_DF;
			dump(2+1);
			return;
		}
		else if(eRear>H_NEAR){				// if the rear requires an up
			if(bAllowUp){
				x.tState = eRear/H_RATE;	// tenths of a second
				x.mState = M_UR;
				pump(8+4);
			}
			else{
				x.mState = M_HOLD;			// no up allowed so wait
				x.tState = 10;
			}
			return;
		}
		else if(eFront>H_NEAR){				// if the front requires an up
			if(bAllowUp){
				x.tState = eFront/H_RATE;	// tenths of a second
				x.mState = M_UF;
				pump(2+1);
			}
			else{
				x.mState = M_HOLD;			// no up allowed so wait
				x.tState = 10;
			}
			return;
		}
		else if(eRear<-H_NEAR){				// if the rear requires a down
			x.tState = -eRear/H_RATE;		// tenths of a second
			x.mState = M_DR;
			dump(8+4);
			return;
		}
		else
			x.step = false;
	}

	// Not a step so we need to assess the struts and pick the greatest error.
	char pErr=0;		// index of strut
	int  err=0;			// max error
	for(char i=0; i<4; i++){
		int e = x.height[i];
		e = e - x.tHeight;		// gives error
		e = abs(e);
		if(e>err){
			err = e;
			pErr = i;
		}
	}
	// anything we need to do?
	if(err<H_NEAR){				// no, so leave the state on M_HOLD...
		x.sMode = x.tMode;		// copy target mode to current mode- we made it
		x.tState = 100;			// ...for ten seconds
		return;
	}
	
	err = x.tHeight - x.height[pErr];	// get back to the signed error
	// positive requires 'up'
	if(err>0){
		if(bAllowUp){
			x.tState = err/H_RATE;			// tenths of a second
			x.mState = M_URF+pErr;
			pump(1<<pErr);
		}
		else{										// no up allowed so wait
			x.mState = M_HOLD;
			x.tState = 10;
		}
		return;
	}
	// else down
	x.tState = -err/H_RATE;				// tenths of a second
	x.mState = M_DRF+pErr;
	dump(1<<pErr);
	return;
}
//------------------------------------------------------------------------------
// This is timer3
//------------------------------------------------------------------------------
// Datasheet page 139
// We prescale the 48MHz CLK/4 by 1 and then count 12000
// giving a 1mSec master clock
#define T3V	(0xffff-12000+1)	// start point to overflow 12000 ticks later

void mSec()
{
	// Reset the counter for another 12000 ticks
	TMR3L = T3V & 0xff;
	TMR3H = T3V >> 8;

	// Update absolute time
	if(++x.mSecs>=1000){
		x.mSecs = 0;
		x.secs++;				// seconds wrap at 18 hours
	}
	// Check for zero RPM
	// rpmTicks is set to zero every time frpm() gets called at spark rate)
	if(x.rpmTicks>50) x.RPM = 0; else x.rpmTicks++;	// 50mS is way too long

	// All the remaining code might run less often than 1mSec but to
	// spread the load out I call them at all sorts of odd numbers
	
	// Run the compressor loop every half a second
	if(x.mSecs==150 || x.mSecs==650)
		compressorTick();

	// Trigger the A2D converter cycle ten times a second
	if(x.mSecs%100==10)
		ADCON0 = 3;				// this starts on A2D 0

	// Run the Ride Height switch every mSec
	rhSwitch();

	// Run the money program 10 times a second
	if(x.mSecs%100==40)
		tilt();					// call the air suspension

	// Count the 'button pressed' timers ten times a second
	if(x.mSecs%100==70){
		if(xUP){
			if(x.upHold<250) 	x.upHold++;
		}else					x.upHold=0;
		if(xDOWN){
			if(x.downHold<250)	x.downHold++;
		}else					x.downHold=0;
	}
	// Do the things that have to be done every second
	// but don't do them all at once
	if(x.mSecs==100)
		flip(xLED1);			// toggle the first LED

	if(x.mSecs==200){
		x.rawMPH = x.mphTicks;
		x.mphTicks = 0;
		// I started with the statement that we get 70 ticks/sec at 50mph
//		x.MPH	= ((int)x.rawMPH * 50) / 70;
		// but that gave 80mph at an indicated 50 so now I use
		x.MPH	= ((int)x.rawMPH * 25) / 56;
	}
	
	if(x.mSecs==300){
		// Drive the compressor timer up and down
		if(x.compON==1 && x.compTimer==COMP_LIMIT)
			x.compON = 2;										// cooldown mode
		else if(x.compON==1 && x.compTimer<COMP_LIMIT)
			x.compTimer++;
		else if(x.compON==0 && x.compTimer>0)
			x.compTimer--;
		else if(x.compON==2 && x.compTimer==COMP_RESET){
			x.compON = 0;										// clear cooldown mode
			x.compTimer--;
		}
		else if(x.compON==2 && x.compTimer>0)
			x.compTimer--;
		
	}
	
	if(x.mSecs==400){	
		// Generate the bHighSpeed signal to shift to motorway mode
		// The LR rules are:
		// Normal goes to Low at speeds over 50mph for >30secs
		// It switches back if speed<35mph for more then 30secs
		if(!x.bHighSpeed){
			if(x.MPH>50){
				if(x.nHighSpeed>30){ x.bHighSpeed = 1; x.step = 1; }
				else				 x.nHighSpeed++;
			}
			else				  	 x.nHighSpeed = 0;
		}
		else{
			if(x.MPH<35){
				if(x.nHighSpeed>30){ x.bHighSpeed = 0; x.step = 1; }
				else			  	 x.nHighSpeed++;
			}
			else				  	 x.nHighSpeed = 0;
		}
	}
}
//------------------------------------------------------------------------------
// Various text utility routines
//------------------------------------------------------------------------------

// Routines to send hex numbers on USB
void usbNibble(unsigned char c)
{
	c &= 0xf;
	usbChar(c>9 ? c+'A'-10 : c+'0');
}
void usbByte(unsigned char c)
{
	usbNibble(c>>4);
	usbNibble(c);
}
void usbWord(unsigned int c)
{
	usbNibble(c>>12);
	usbNibble(c>>8);
	usbNibble(c>>4);
	usbNibble(c);
}
// Send a text string on USB
void usbText(char *p)
{
	while(*p) usbChar(*p++);
}
// find a character in a string
char *strchr(char *p, char c)
{
	while(*p){
		if(*p==c) return p;
		p++;
	}
	return 0;
}
// decode two hex characters as ports/bit and toggle it
void toggle(unsigned char* p)
{
	int pt = *p++-'0';
	int bt  = *p-'0';
	if(pt>=0 && pt<=4 && bt>=0 && bt<=7)
		flip(pt, bt);
}
//------------------------------------------------------------------------------
// memDump()
//------------------------------------------------------------------------------

// x.nDump is the index into x of how far we got.
// We set x.nDump to 0 when we finish.
// OK... 8 bytes may have gone as D XXXX\r\n so we have 56 bytes to play with.
// Sending hex the best system seems to be to send 16 bytes as XX then \r\n
// using 34 bytes every time we are called until x.nDump>sizeof(x)

void memDump()
{
	unsigned int n = sizeof(x);				// total RAM to dump
	if(x.nDump>=n){ x.nDump=0; return; }	// safety trap
	char *p = (char*)&x.RAM_START;			// start of structure
	unsigned int i = n-x.nDump;				// distance to go
	if(i>16) i=16;							// how much we will do
	for(unsigned int j=0; j<i; j++)
		usbByte(p[x.nDump+j]);
	usbText("\r\n");
	x.nDump += i;
	if(x.nDump>=n) x.nDump=0;
}
//------------------------------------------------------------------------------
// General processing. Basically this is user interface as everything else
// runs on the timer ticks to give real time.
//------------------------------------------------------------------------------

void UserLoop()
{
	unsigned char *p1, *p2;
	out(xLED2, x.bHold);

	// if the RAM looks corrupt restart the world
	if(x.RAM_START!='S' || x.RAM_END!='F'){
		usbText("GOODBYE CRUEL WORLD");
		die();
	}

	// there is a 64 byte limit to the USB output and, although we can be
	// called every mS to send another we need to wait for it
	if(cdc_trf_state==CDC_TX_READY){
		
		// firstly do we have a memory dump running?
		if(x.nDump)
			memDump();
		
		// Commands:
		// d	ram dump
		// h	run reset to start condition
		// p	report version and build
		// s	save current data to eeRom
		// t	toggle a bit: use as tHH where first is port second is bit
		// u	clear errors
		// z	calibrate to bump-stops
		else if(x.usbCmd[0] && strchr(x.usbCmd, '\r')){	// if there is a '\r'
			// remember the 64 byte output buffer...
			switch(x.usbCmd[0]){
			case 'd':
				normalise();
				x.vDebug1 = xROLL && xENGINE;	// condition to use lag
				usbText("M ");
				x._inputs[0] = PORTA;
				x._inputs[1] = PORTB;
				x._inputs[2] = PORTC;
				x._inputs[3] = PORTD;
				x._inputs[4] = PORTE;
				p1 = &x.RAM_START;
				p2 = &x.RAM_END;
				usbWord((int)p2-(int)p1+1);
				usbText("\r\n");
				memDump();
				break;
						
			case 'h':
				UserInitialise();			// reset everything
				goto cr;
	
			case 'p':						// ping : 'are you there?'
				usbText("Version " VERSION ", build " __TIMESTAMP__
	#ifdef N_DEBUG
				" DEBUG"
	#endif
				"\r\n");
				break;
	
	
			case 's':
				if(x.eeAddr==0)
					eeWrite();
				goto ok;
	
			case 't':						// toggle a bit
				toggle(x.usbCmd+1);
				goto ok;
				
			case 'u':						// clear errors
				x.bugs = 0;
				for(int i=0; i<sizeof(x.eeRom_errors); i++)
					x.eeRom_errors[i] = 0;
				goto cr;
			
			case 'z':
				calibrate();
	ok:			usbText("OK");
	cr:			usbText("\r\n");
				break

			default:
				usbText("?\r\n");
				break;
			}
			// Clear the buffer
			x.usbCmd[0]=0;
			x.usbNcmd=0;
		}
	}
}
